#include "_buffer.h"
#include "_stralloc.h"
#include "str.h"
#include "ftplib.h"
#include "bailout.h"
#include "ftpcopy.h"
#include "error.h"
#include "timeoutio.h"
#include "doconnect.h"
#include "_readwrite.h"
#include "iopause.h"
#include "close.h"

static stralloc io_i_mem;
static stralloc io_o_mem;

MKTIMEOUTREAD(30)
MKTIMEOUTWRITE(30)


char *
ccread(void)
{
	static stralloc sa=STRALLOC_INIT;
	switch(ftp_cc_read(&io_i,&sa)) {
	case  0: errno=0; /* FALL THROUGH */
	case -1: return 0;
	}
	return sa.s;
}
void
cmdwrite2(const char *s1, const unsigned char *s2)
{
	if (-1==ftp_cc_write_cmd_ss(&io_o,s1,s2))
			bailout(errno,"remote write",0,0,0);
}
void x2(const char *where)
{
	char *p;
	p=ccread();
	if (!p && !errno) bailout(0,"EOF reading from remote",0,0,0);
	if (!p) bailout(errno,"read (",where,")",0);
	if (*p != '2') bailout(0,where,": ",p,0);
}
void sx2(const char *what)
{
	cmdwrite1(what);
	x2(what);
}
void
cmdwrite1(const char *s1)
{
	if (-1==ftp_cc_write_cmd_s(&io_o,s1)) bailout(errno,"remote write",0,0,0);
}
#define LOGSTR(x) if (-1==buffer_puts(buffer_1,(x)))  \
	bailout(errno,"can't write to log",0,0,0);
#define LOGMEM(x,y) if (-1==buffer_put(buffer_1,(x),(y)))  \
	bailout(errno,"can't write to log",0,0,0);
#define LOGFLUSH() \
	if (-1==buffer_flush(buffer_1)) bailout(errno,"can't flush log",0,0,0);

void log1(const char *s1) { LOGSTR(s1); LOGFLUSH(); }
void log2(const char *s1, const char *s2)
	{ LOGSTR(s1); LOGSTR(s2); LOGFLUSH(); }
void log3(const char *s1, const char *s2, const char *s3)
	{ LOGSTR(s1); LOGSTR(s2); LOGSTR(s3); LOGFLUSH(); }
void log4(const char *s1, const char *s2, const char *s3, const char *s4)
	{ LOGSTR(s1); LOGSTR(s2); LOGSTR(s3); LOGSTR(s4); LOGFLUSH(); }
void logmem(const char *s1,unsigned int l) { LOGMEM(s1,l); LOGFLUSH(); }

int
do_pasv(void)
{
	int x;
	x=ftp_cc_pasv(&io_i,&io_o);
	switch(x) {
	case 0: bailout(0,"EOF during PASV",0,0,0);
	case -1: bailout(errno,"IO error during PASV",0,0,0);
	case -2: bailout(0,"parse error during PASV",0,0,0);
	}
	return x;
}
int
connect_auth (const char *host, const char *o_user, 
	const char *o_pass, const char *o_acct, int tries)
{
	if (!stralloc_ready(&io_i_mem,BUFFER_INSIZE)) oom();
	if (!stralloc_ready(&io_o_mem,BUFFER_OUTSIZE)) oom();

	while (1) {
		int need_auth = 1;
		int need_acct = 0;
		char *p;
		int sock;
		if (!tries) 
			bailout(0,"cannot connect / log in",0,0,0);
		tries--;
		sock=doconnect(host);
		if (sock==-1) {
			if (!tries) bailout(errno,"connect",0,0,0);
			goto dosleep1;
		}

		buffer_init(&io_i,(BUFFER_CAST)TIMEOUTREADFN(30),sock,
		io_i_mem.s,BUFFER_INSIZE);
		buffer_init(&io_o,(BUFFER_CAST)TIMEOUTWRITEFN(30),sock,
		io_o_mem.s,BUFFER_OUTSIZE);
											
		p = ccread ();
		if (!p) {
			if (tries) goto dosleep;
			if (!errno) bailout(0,"EOF reading greeting",0,0,0);
			bailout(errno,"read (greeting)",0,0,0);
		}
		if (o_loglevel > 2)
			log2 (p, "\n");
		if (!str_start (p, "220 ")) {
			if (!tries) bailout (0, "greeting message: ", p, 0, 0);
			goto dosleep;
		}
		if (!o_user || !*o_user) 
			need_auth=0;
		if (str_start(p,"220 Features:")) {
			char *q;
			for (q = p; *q; q++) {
				if (*q == ' ' && q[1] == 'a')
					break;
			}
			if (*q)
				if (str_equal (o_user, "anonymous")
					|| str_equal (o_user, "ftp"))
					need_auth = 0;
		}

		if (need_auth) {
			cmdwrite2 ("USER ", o_user);
			p = ccread ();
			if (!p) {
				if (tries) goto dosleep;
				if (!errno) bailout(0,"EOF reading USER answer",0,0,0);
				bailout(errno,"read (USER answer)",0,0,0);
			}
			if (*p == '5') {
				if (tries) goto dosleep;
				bailout (0, "USER response: ", p, 0, 0);
			}
			if (*p == '2')
				return sock;
			if (o_pass && *o_pass) {
				cmdwrite2 ("PASS ", o_pass);
				p = ccread ();
				if (!p) {
					if (tries) goto dosleep;
					if (!errno) bailout(0,"EOF reading PASS answer",0,0,0);
					bailout(errno,"read (PASS answer)",0,0,0);
				}
				if (str_start(p,"332"))
					need_acct=1;
				else if (*p != '2') {
					if (!tries) 
						bailout(0,"password rejected: ",p,0,0);
					goto dosleep;
				}
			}
			if (need_acct) {
				if (!o_acct)
					bailout(0,"server asked for ACCT.",0,0,0);
				cmdwrite2 ("ACCT ", o_acct);
				p = ccread ();
				if (!p) {
					if (tries) goto dosleep;
					if (!errno) bailout(0,"EOF reading ACCT answer",0,0,0);
					bailout(errno,"read (ACCT answer)",0,0,0);
				}
				if (*p != '2') {
					if (!tries) 
						bailout(0,"account rejected: ",p,0,0);
					goto dosleep;
				}
			}
		}
		return sock;
	  dosleep:
		close(sock);
	  dosleep1:
	  	{
			iopause_fd x;
			struct taia now;
			struct taia deadline;
			taia_now(&now);
			taia_uint(&deadline,5);
			taia_add(&deadline,&now,&deadline);
			iopause(&x,0,&deadline,&now);
		}
	}
}
