#include <unistd.h>
#include <stdlib.h>
#include <dirent.h>
#include <stdio.h> /* rename */
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include "ftpparse.h"
#include "_buffer.h"
#include "_stralloc.h"
#include "_scan.h"
#include "_ip4.h"
#include "_alloc.h"
#include "byte.h"
#include "getln.h"
#include "socket.h"
#include "ftplib.h"
#include "doconnect.h"
#include "str2num.h"
#include "strhash.h"
#include "str_ulong.h"
#include "uogetopt.h"
#include "uotime.h"
#include "uo_wildmat.h"
#include "urlparse.h"
#include "timeoutio.h"
#include "iopause.h"
#include "ftpcopy.h"
#include "bailout.h"
#include "str.h"
#include "error.h"
#include "open.h"
#include "have_futimes.h"


buffer io_i;
buffer io_o;
static buffer io_d;
static stralloc io_d_mem;
int data_sock=-1;
static strhash_t hash_ids;
static strhash_t hash_fns;
static const char *o_user;
static const char *o_pass;
static const char *o_acct;
unsigned long o_loglevel=1;
static int o_do_delete=1;
static int o_symlink_hack=0;
static int o_directories_only=0;
static int o_remove_dir=0;
static const char *o_list_options=0;
static stralloc o_exclude=STRALLOC_INIT;
static unsigned long o_max_days=0;
static unsigned long o_tries=1;

MKTIMEOUTREAD(30)

/* static void logmem(char *s, size_t len) { LOGMEM(s,len); LOGFLUSH(); } */

static int do_symlink(const char *f,const char *t)
{
	static stralloc s=STRALLOC_INIT;
	size_t l1=0;
	size_t slash=0;
	while (f[l1] && t[l1]==f[l1]) {
		if (t[l1]=='/') slash=l1;
		l1++;
	}
	if (!f[l1] && t[l1]=='/')
		slash=l1;

	if (slash++) {
		const char *p=t+slash+1;
		s.len=0;
		while (*p) {
			if (*p=='/') 
				if (!stralloc_cats(&s,"../")) oom();
			p++;
		}
		if (!stralloc_cats(&s,f+slash)) oom();
		if (s.len==0) if (!stralloc_cats(&s,".")) oom();
		if (!stralloc_0(&s)) oom();
		return symlink(s.s,t);
	}
	if (!stralloc_cats(&s,f) || !stralloc_0(&s)) oom();
	return symlink(s.s,t);
}

static stralloc *
canon(stralloc *l_dir) /* feed me with 0 terminated strings only */
{
	static stralloc c=STRALLOC_INIT;
	char *p,*end,*w;
	if (!stralloc_copy(&c,l_dir)) oom();
	if (!stralloc_readyplus(&c,l_dir->len+1)) oom();
	p=l_dir->s;
	end=l_dir->s+l_dir->len;
	w=c.s;

	while (*p) {	
		if (p[0]=='/' && p[1]=='.' && p[2]=='.' && (!p[3] || p[3]=='/')) {
			while (w!=c.s && w[-1]!='/') w--;
			*w='/';
			p+=4;
		} else if (p[0]=='/' && p[1]=='/')
			p++;
		else
			*w++=*p++;
	}
	c.len=w-c.s;
	if (!stralloc_0(&c)) oom();
	return &c;
}

static void
hash_it(struct ftpparse *x, stralloc *l_dir, int id_also)
{
	stralloc *t;
	t=canon(l_dir);

	if (1!=strhash_enter(&hash_fns,1,t->s,t->len,0,0,0))
		oom();

	if (!x->idlen || !id_also) {
		if (1!=strhash_enter(&hash_ids,
				1,t->s,t->len,
				1,t->s,t->len)) 
			oom();
	} else {
		if (1!=strhash_enter(&hash_ids,
				1,x->id,x->idlen,
				1,t->s,t->len)) 
			oom();
	}
}

static int
download(struct ftpparse *x, stralloc *r_dir, stralloc *l_dir)
{
	char *p;
	int fd;
	const char *e;
	const char *slash;
	const char *cc;
	static stralloc tmpfn=STRALLOC_INIT;
	buffer save;
	static stralloc savemem=STRALLOC_INIT;
	struct stat st;
	char *fptr;
	unsigned int flen;
	int found;
	struct timeval timval[2];

	e=l_dir->s+l_dir->len;
	for (cc=slash=l_dir->s;cc!=e;cc++)
		if (*cc=='/')
			slash=cc;
	if (*slash=='/') slash++;
	if (!stralloc_copys(&tmpfn,".tmp.")
		|| !stralloc_catb(&tmpfn,slash,e-slash)
		|| !stralloc_0(&tmpfn)) oom();
	unlink(tmpfn.s);

	found=strhash_lookup(&hash_ids,x->id,x->idlen,&fptr,&flen);

	if (0==lstat(l_dir->s,&st)) {
		if (st.st_mtime==x->mtime && st.st_size==x->size) {
			hash_it(x,l_dir,!found);
			if (o_loglevel > 1)	
				log2(l_dir->s, ": identical file found\n");
			return 1;
		}
		if (o_loglevel > 2)	{
			char nb[STR_ULONG];
			log2(l_dir->s, ": lfacts: ");
			str_ulong(nb,st.st_mtime);
			log2(nb," ");
			str_ulong(nb,st.st_size);
			log2(nb,"\n");
		}
	}
	
	if (found) {
		if (0==link(fptr,tmpfn.s)
			|| 0==do_symlink(fptr,tmpfn.s))  {
			if (-1==rename(tmpfn.s,l_dir->s)) {
				int er=errno;
				unlink(tmpfn.s);
				bailout(er,"rename", tmpfn.s," -> ",l_dir->s);
			}
			if (o_loglevel > 1)
				log4(l_dir->s, " (sym)linked to ",fptr,"\n");
			hash_it(x,l_dir,1);
			return 1;
		}
	}
	if (o_directories_only) {
		hash_it(x,l_dir,1);
		return 0;
	}
	if (o_max_days && x->mtime+o_max_days*86400<uo_now()) {
		hash_it(x,l_dir,1);
		if (o_loglevel > 1)	
			log2(l_dir->s, ": too old\n");
		return 1;
	}

	if (data_sock==-1)
		data_sock=do_pasv();

	fd=open_trunc_mode(tmpfn.s,0644);
	if (fd==-1) bailout(errno,tmpfn.s,": open",0,0);
	if (!stralloc_ready(&savemem,BUFFER_OUTSIZE)) oom();
	buffer_init(&save,(BUFFER_CAST)write,fd,savemem.s,BUFFER_OUTSIZE);


	cmdwrite2("RETR ",r_dir->s);
	p=ccread();
	if (!p) bailout(errno,"read (RETR answer)",0,0,0);
	if (*p!='1') {
		warning(0,"RETR ",r_dir->s,": ",p);
		unlink(tmpfn.s);
		close(fd);
		return 0;
	}
	buffer_init(&io_d,(BUFFER_CAST)TIMEOUTREADFN(30),data_sock,
		io_d_mem.s,BUFFER_INSIZE);
	while (1) {
		int l;
		char *q;
		l=buffer_feed(&io_d);
		if (l==-1) bailout(errno,"remote read",0,0,0);
		if (l==0) break;
		q=buffer_peek(&io_d);
		if (-1==buffer_put(&save,q,l)) bailout(errno,tmpfn.s,": write",0,0);
		buffer_seek(&io_d,l);
	}
	close(data_sock);
	x2("RETR finish");
	if (buffer_flush(&save)) bailout(errno,tmpfn.s,": write",0,0);
	data_sock=-1;
	if (-1==fsync(fd)) bailout(errno,tmpfn.s,": fsync",0,0);

#ifdef HAVE_FUTIMES
	timval[0].tv_usec=timval[1].tv_usec=0;
	timval[0].tv_sec=timval[1].tv_sec=x->mtime;
	if (-1==futimes(fd,timval))
		warning(errno,tmpfn.s,": futimes()",0,0);
#endif
	if (-1==close(fd)) bailout(errno,tmpfn.s,": close",0,0);

#ifndef HAVE_FUTIMES
	/* it seems reiserfs on linux didn't honor utimes() if it was 
	 * called before the close(). */
	timval[0].tv_usec=timval[1].tv_usec=0;
	timval[0].tv_sec=timval[1].tv_sec=x->mtime;
	if (-1==utimes(tmpfn.s,timval))
		warning(errno,tmpfn.s,": utimes()",0,0);
#endif

	if (-1==rename(tmpfn.s,l_dir->s)) {
		int er=errno;
		unlink(tmpfn.s);
		bailout(er,"rename", tmpfn.s," -> ",l_dir->s);
	}
	if (o_loglevel)
		log2(l_dir->s, ": download successful\n");
	hash_it(x,l_dir,1);

	return 1;
}

static void
remove_dir (stralloc *s)
{
	DIR *d;
	d=opendir(s->s);
	if (!d) bailout(errno,"cannot opendir ",s->s,0,0);
	s->len--;
	if (!stralloc_append(s,"/")) oom();
	while (1) {
		struct dirent *e;
		int pos=s->len;
		e=readdir(d);
		if (!e) break;
		if (e->d_name[0]=='.') {
			if (e->d_name[1]=='.' && e->d_name[2]=='\0') continue;
			if (e->d_name[1]=='\0') continue;
		}
		if (!stralloc_cats(s,e->d_name)) oom();
		if (!stralloc_0(s)) oom();
		if (-1==unlink(s->s)) {
			if (errno==error_isdir) 
				remove_dir(s);
			else
				bailout(errno,"cannot unlink ",s->s,0,0);
		}
		s->len=pos;
	}
	s->s[s->len++]=0;
}

static int 
loop(stralloc *r_dir, stralloc *l_dir)
{
	char *p,*e;
	stralloc dirdata=STRALLOC_INIT;

	if (data_sock==-1)
		data_sock=do_pasv();

	cmdwrite2("CWD ",r_dir->s);

	p=ccread();
	if (!p) bailout(errno,"read (CWD answer)",0,0,0);
	if (*p!='2') 
		return 0;

	if (o_list_options)
		cmdwrite2("LIST ", o_list_options);
	else
		cmdwrite1("LIST");
	p=ccread();
	if (!p) bailout(errno,"read (LIST answer)",0,0,0);
	if (*p!='1')
		return 0;

	if (-1==chdir(l_dir->s)) {
		if (-1==mkdir(l_dir->s,0755)) 
			bailout(errno,"mkdir ",l_dir->s,0,0);
		if (-1==chdir(l_dir->s)) bailout(errno,"chdir ",l_dir->s,0,0);
	}

    if (-1==ftp_read_list(data_sock,&dirdata))
		bailout(errno,"cannot read directory",0,0,0);

	close(data_sock);
	data_sock=-1;
	x2("LIST");

	p=dirdata.s;
	e=dirdata.s+dirdata.len;
	while (p!=e) {
		struct ftpparse x;
		size_t l=str_len(p);
		if (0==ftpparse(&x,p,l)) {
			log3("cannot parse LIST line: ",p,"\r\n");
		} else if (x.name[0]=='.'
			&& (x.namelen==1 || (x.namelen==2 && x.name[1]=='.'))) {
			if (o_loglevel > 1) {
				log2(l_dir->s,"/");
				logmem(x.name,x.namelen);
				log1(": ignored\n");
			}
		} else if (x.name[byte_chr(x.name,x.namelen,'/')] == '/') {
			/* file name with a slash in it: no good.
			 * Can't happen and has security implications. */
			log2(l_dir->s,"/");
			logmem(x.name,x.namelen);
			log1(": ignored\n");
		} else {
			size_t rpos=r_dir->len;
			size_t lpos=l_dir->len;
			int done=0;
			stralloc *ca;
			if (o_loglevel > 2) {
				char nb1[STR_ULONG];
				char nb2[STR_ULONG];
				str_ulong(nb1,x.mtime);
				str_ulong(nb2,x.size);
				log2(l_dir->s,"/");
				logmem(x.name,x.namelen);
				log4(": facts: ",nb1," ",nb2); log1("\n");
			}
			
			r_dir->s[r_dir->len-1]='/';
			if (!stralloc_catb(r_dir,x.name,x.namelen)
				|| !stralloc_0(r_dir)) oom();
			ca=canon(r_dir);
			if (o_exclude.len) {
				char *end=o_exclude.s+o_exclude.len;
				char *ptr=o_exclude.s;
				char *info=0; /* keep gcc quiet */
				int exclude=0;
				while (ptr!=end) {
					int flag=(*ptr=='-');
					ptr++;
					if (uo_wildmat(ptr,ca->s,ca->len-1)) {
						exclude=flag;
						if (flag)
							info=ptr;
					}
					ptr+=str_len(ptr)+1;
				}
				if (exclude) {
					done=1;
					if (o_loglevel > 1) {
						write(1,ca->s,ca->len-1); 
						log3(": excluded, matched `",info,"'\n");
					}
				}
			}
			if (!done && l>4 && *p=='l' && o_symlink_hack) {
				char *q;
				char *end;
				q=p;
				end=q+l-4;
				while (q!=end) {
					if (*q==' ' && q[1]=='-'
						&& q[2]=='>' && q[3]==' ')
						{ q+=4 ; break; }
					q++;	
				}
				if (q!=end) {
					stralloc t=STRALLOC_INIT;
					stralloc idstr=STRALLOC_INIT;
					if (!stralloc_copy(&t,l_dir)) oom();
					t.s[t.len-1]='/';
					if (!stralloc_catb(&t,q,str_len(q)) || !stralloc_0(&t)) 
						oom();
					if (o_loglevel > 1) {
						log1("symlink '"); 
						write(1,t.s,t.len-1); 
						log1("'\n");
					}
					if (!stralloc_copy(&idstr,canon(&t))) oom();
					if (!stralloc_copy(&t,l_dir)) oom();
					t.s[t.len-1]='/';
					if (!stralloc_catb(&t,x.name,x.namelen)) oom();
					if (!stralloc_0(&t)) oom();
					unlink(t.s);
					rmdir(t.s);
					if (0!=do_symlink(idstr.s,t.s))  {
						if (errno==error_isdir) 
							if (o_remove_dir)
								remove_dir(&t);
						if (0!=do_symlink(idstr.s,t.s))
							bailout(errno,"symlink ",idstr.s, 
								" -> ",t.s);
					}
					hash_it(&x,&t,1);
					stralloc_free(&t);
					stralloc_free(&idstr);
					done=1;
				}
			}
			if (x.flagtrycwd && !done) {
				int found;
				char *fptr;
				unsigned int flen;

				l_dir->s[l_dir->len-1]='/';
				if (!stralloc_catb(l_dir,x.name,x.namelen)
					|| !stralloc_0(l_dir)) oom();
				found=strhash_lookup(&hash_ids,x.id,x.idlen,&fptr,&flen);
				if (found) {
					if (o_loglevel > 1)
						log4(l_dir->s,": (sym)linking, ID identical to `",
							fptr,"'\n");
					unlink(l_dir->s);
					rmdir(l_dir->s);
					if (0!=do_symlink(fptr,l_dir->s)) 
						bailout(errno,"symlink ",fptr, " -> ",l_dir->s);
					hash_it(&x,l_dir,1);
					done=1;
				} else {
					/* hash it now: "dir" -> "." symlinks */
					hash_it(&x,l_dir,1);

					if (1==loop(r_dir, l_dir)) {
						struct timeval timval[2];
						timval[0].tv_sec=x.mtime;
						timval[0].tv_usec=x.mtime;
						timval[1].tv_sec=x.mtime;
						timval[1].tv_usec=x.mtime;
						if (-1==utimes(l_dir->s,timval))
							warning(errno,l_dir->s,": utimes()",0,0);
						done=1;
					}
				}
				l_dir->len=lpos;
				l_dir->s[l_dir->len-1]='\0';
			}
			if (x.flagtryretr && !done) {
				l_dir->s[l_dir->len-1]='/';
				if (!stralloc_catb(l_dir,x.name,x.namelen)
					|| !stralloc_0(l_dir)) oom();
				if (1==download(&x,r_dir, l_dir))
					done=1;
				l_dir->len=lpos;
				l_dir->s[l_dir->len-1]='\0';
			}
			r_dir->len=rpos;
			r_dir->s[r_dir->len-1]='\0';
		}
		p+=l+1;
	}
	stralloc_free(&dirdata);
	return 1; /* done */
}

static int 
initialentity(struct ftpparse *fp, stralloc *r_dir)
{
	int count=0;
	char *p;
	static stralloc dirdata=STRALLOC_INIT; /* we return pointers into that mem */

	if (data_sock==-1)
		data_sock=do_pasv();

	cmdwrite2("LIST ",r_dir->s);
	p=ccread();
	if (!p) bailout(errno,"read (LIST answer)",0,0,0);
	if (*p!='1') bailout(0,"initial list response: ",p,0,0);

    count=ftp_read_list(data_sock,&dirdata);
	if (-1==count)
		bailout(errno,"cannot read directory",0,0,0);
	close(data_sock);
	data_sock=-1;
	x2("LIST");
	if (count==1) {
		stralloc s=STRALLOC_INIT;
		if (!stralloc_catb(&s,dirdata.s,str_len(dirdata.s))) oom();
		if (0==ftpparse(fp,s.s,s.len)) {
			fp->flagtrycwd=1;
			fp->flagtryretr=1;
		}
	} else if (count>1) {
		fp->flagtrycwd=1; 
		fp->flagtryretr=0;
	}

	return 1;
}



static void
delete(stralloc *dn)
{
	DIR *d;
	size_t pos;
	d=opendir(dn->s);
	if (!d) bailout(errno,dn->s,": opendir",0,0);
	dn->len--;
	pos=dn->len;
	while (1) {
		struct dirent *e;
		struct stat st;
		int found;

		stralloc *t;
		dn->len=pos;
		e=readdir(d);
		if (!e) 
			break;
		if (e->d_name[0]=='.'
			&& (!e->d_name[1]
				|| (e->d_name[1]=='.' && e->d_name[2]==0)))
			continue;
		if (!stralloc_append(dn,"/")) oom();
		if (!stralloc_cats(dn,e->d_name)) oom();
		if (!stralloc_0(dn)) oom();
		if (-1==lstat(dn->s,&st))
			continue;
		if (S_ISDIR(st.st_mode)) {
			delete(dn);
			continue;
		}
		t=canon(dn);
		found=strhash_lookup(&hash_fns,t->s,t->len,0,0);
		if (!found) {
			if (-1==unlink(dn->s) && -1==rmdir(dn->s))
				bailout(errno,dn->s,": unlink",0,0);
			if (o_loglevel)
				log2(dn->s,": not found on remote, deleted\n");
		} else {
			if (o_loglevel>3)
				log2(dn->s,": found on remote\n");
		}
	}
	closedir(d);
}

static void 
usage(void) 
{
	bailout(0,
	  "usage: ftpcopy [options] host[:port] remotedir [localdir]\n"
	  "   or: ftpcopy [options] ftp://host[:port]/remotedir [localdir]\n"
	  "  use the --help option to get a description of the options"
	  ,0,0,0);
}

static void
callback_exclude(uogetopt_t *g, const char *s)
{
	(void) g;
	if (!stralloc_catb(&o_exclude,"-",1)
		|| !stralloc_cats(&o_exclude,s)
		|| !stralloc_0(&o_exclude)) oom();
}
static void
callback_include(uogetopt_t *g, const char *s)
{
	(void) g;
	if (!stralloc_catb(&o_exclude,"+",1)
		|| !stralloc_cats(&o_exclude,s)
		|| !stralloc_0(&o_exclude)) oom();
}

static uogetopt_t myopts[] =
{
{  0, "account", UOGO_STRING, &o_acct,0,
"Send ACCOUNT as account name during login phase.\n"
"Note: this is _not_ the user name (-u), but the\n"
"name of what could be called a subaccount\n"
"implemented by a few servers.\n","ACCOUNT"},
{'d', "directories-only", UOGO_FLAG, &o_directories_only,1,
"Only create the directory hierarchie.\n"
"Do not download files.\n"
"Any file in the tree will be deleted unless\n"
"the -n option is also given.",0},
{'l', "loglevel", UOGO_ULONG, &o_loglevel,0,
"Controls the amount of logging done.\n"
"0: nothing except warnings and error messages.\n"
"1: downloads and deletes.\n"
"2: links/symlinks created, files we already got.\n"
"3: useless stuff.",0},
{'L', "list-options", UOGO_STRING, &o_list_options,1,
"Add OPTS to LIST command.\n"
"This allows to pass arbitrary options to the\n"
"FTP servers LIST command. Note that ftpcopy does\n"
"not cope well with recursive directory listings.","OPTS"},
{'m', "max-days", UOGO_ULONG, &o_max_days,0,
"Restrict on modification time.\n"
"Download only files modified in the last\n"
"MAX days.","MAX"},
{'n', "no-delete", UOGO_FLAG, &o_do_delete,0,
"Do not delete files.\n"
"This influences the cleanup step, it doesn't\n"
"stop ftpcopy from deleting files during the\n"
"downloads.",0},
{'p', "pass", UOGO_STRING, &o_pass,0,
"Use PASS as password to login on the ftp server.\n"
"The default is `anonymous@invalid.example'.","PASSWORD"},
{'R', "remove-dir", UOGO_FLAG, &o_remove_dir,1,
/*2345678901234567890123456789012345678901234567890 */
"Remove directories `in the way'.\n"
"If a directory is in our way because it has\n"
"changed into something different in the remote\n"
"site, then remove that directory recursively.\n"
"This may be the default in a few releases.",0},
{'\0', "keep-dir", UOGO_FLAG, &o_remove_dir,0,
/*2345678901234567890123456789012345678901234567890 */
"Don't remove directories `in the way'.\n"
"Keep directories, even it they are blocking our\n"
"operation. Note that ftpcopy will _not_ continue\n"
"in that case. This option is currently the\n"
"default, but that may change.",0},
{'s', "symlink-hack", UOGO_FLAG, &o_symlink_hack,1,
"Deal with symbolic links.\n"
"Only useful to mirror sites which create listings\n"
"through /bin/ls. Will fail if a file name in a\n"
"link contains a ` -> ' sequence.",0},
{'\0', "tries", UOGO_ULONG, &o_tries,0,
"Number of tries to connect and log in.\n"
"The default is 1.",0},
{'u', "user", UOGO_STRING, &o_user,0,
"Use NAME to login on the ftp server.\n"
"The default is `anonymous'. Use an empty name\n"
"to avoid a login.","NAME"},
{'x', "exclude", UOGO_CALLBACK, callback_exclude,0,
"Exclude paths matching WILDCARD.\n"
"If WILDCARD matches the full path of the remote\n"
"file then the file will not be downloaded.\n"
/*2345678901234567890123456789012345678901234567890 */
"Use this option more than once to exclude multiple\n"
"patterns.","WILDCARD"},
{'i', "include", UOGO_CALLBACK, callback_include,0,
"Include paths matching WILDCARD.\n"
"If WILDCARD matches the full path of the remote\n"
"file then the file will be downloaded.\n"
"Use this option more than once to exclude multiple\n"
"patterns.\n"
"If both includes and excludes are used then the\n"
"last matching one will be honored. The list starts\n"
"with an implicit '--include *'","WILDCARD"},
{0,"",UOGO_MINARGS | UOGO_HIDDEN,0, 2 , 0,0},
{0,"",UOGO_MAXARGS | UOGO_HIDDEN,0, 4 , 0,0},
{0, 0, 0, 0, 0, 0, 0}           /* --help and --version */
};

int 
main(int argc, char **argv)
{
	struct ftpparse fp;
	const char *host,*remotedir,*localdir=0;
	stralloc d1=STRALLOC_INIT;
	stralloc d2=STRALLOC_INIT;
	stralloc proto=STRALLOC_INIT;
	stralloc user=STRALLOC_INIT;
	stralloc pass=STRALLOC_INIT;
	stralloc hostport=STRALLOC_INIT;
	stralloc rest=STRALLOC_INIT;

	bailout_progname(argv[0]);

	if (!stralloc_ready(&io_d_mem,BUFFER_INSIZE)) oom();

	uogetopt (flag_bailout_log_name, PACKAGE, VERSION, 
		&argc, argv, uogetopt_out, 
	  "usage: ftpcopy [options] host[:port] remotedir [localdir]\n"
	  "   or: ftpcopy [options] ftp://host[:port]/remotedir [localdir]\n",
	  myopts,
	  "  localdir defaults to '.'. If it's not given then the -n option\n"
	  "  must be used.\n"
	  "\nReport bugs to <uwe-ftpcopy@bulkmail.ohse.de>");
	if (urlparse(argv[1],&proto,&user,&pass,&hostport,&rest)) {
		if (!stralloc_0(&proto)) oom();
		if (!str_equal(proto.s,"ftp")) 
			bailout(0,"unsupported URL type",0,0,0);
		if (!hostport.len) bailout(0,"empty host in url",0,0,0);
		if (!stralloc_0(&hostport)) oom();
		host=hostport.s;
		if (!rest.len) if (!stralloc_append(&rest,"/")) oom();
		if (!stralloc_0(&rest)) oom();
		remotedir=rest.s;
		if (!o_user && user.len) {
			if (!stralloc_0(&user)) oom();
			o_user=user.s;
		}
		if (!o_pass && pass.len) {
			if (!stralloc_0(&pass)) oom();
			o_pass=pass.s;
		}
		if (argv[2]) {
			localdir=argv[2];
			if (argv[3]) usage();
		}
	} else {
		if (!argv[2])
			usage();
		host=argv[1];
		remotedir=argv[2];
		if (argv[3])
			localdir=argv[3];
	}
	if (!localdir) {
		char dotdir[]=".";
		if (o_do_delete)
			xbailout(111,0,
				"default `localdir' (.) not allowed without the -n option",
				0,0,0);
		localdir=dotdir;
	}
	if (!o_user) o_user="anonymous";
	if (!o_pass) o_pass="anonymous@example.invalid";

	if (-1==strhash_create(&hash_ids,16,32,strhash_hash)) oom();
	if (-1==strhash_create(&hash_fns,16,32,strhash_hash)) oom();

	if (!stralloc_copys(&d1,remotedir)) oom();
	if (!stralloc_0(&d1)) oom();
	if (*localdir=='/') {
		if (!stralloc_copys(&d2,localdir)) oom();
		if (!stralloc_0(&d2)) oom();
	} else {
		char *buf=alloc(1024);
		if (!buf) oom();
		if (!getcwd(buf,1024)) bailout(errno,"getcwd",0,0,0);
		if (!stralloc_copys(&d2,buf)) oom();
		if (!stralloc_append(&d2,"/")) oom();
		if (!stralloc_cats(&d2,localdir)) oom();
		if (!stralloc_0(&d2)) oom();
		alloc_free(buf);
	}
	
	connect_auth(host,o_user,o_pass,o_acct,o_tries);

	sx2("TYPE I");
	memset(&fp,0,sizeof(fp));
	initialentity(&fp,&d1);
	if (fp.flagtryretr) {
		struct stat st;
		if (-1==chdir(d2.s)) {
			char *p=d2.s+str_rchr(d2.s,'/');
			if (p!=d2.s+d2.len) {
				*p=0;
				if (-1==chdir(d2.s))
					bailout(errno,d2.s,": chdir",0,0);
				*p='/';
			} 
		}
		if (0==stat(d2.s,&st) && S_ISDIR(st.st_mode)) {
			char *q=d1.s+str_rchr(d1.s,'/');
			if (q!=d1.s+d1.len) {
				d2.len--;
				if (!stralloc_cats(&d2,q) || !stralloc_0(&d2)) oom();
			}
		}
		if (download(&fp,&d1,&d2))
			exit(0);
	}
	if (fp.idlen)
		/* for the weird symlinks to the servers root */
		hash_it(&fp,&d2,1);
	if (fp.flagtrycwd)
		if (loop(&d1,&d2)) {
			if (o_do_delete)
				delete(&d2);
			exit(0);
		}
	exit(1);
}
