#include <unistd.h>
#include <netinet/in.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <dirent.h>
#include <stdio.h> /* rename */
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include "_buffer.h"
#include "_stralloc.h"
#include "_ip4.h"
#include "getln.h"
#include "ftplib.h"
#include "doconnect.h"
#include "str2num.h"
#include "str.h"
#include "fmt.h"

static int
ftp_cc_write_escaped(buffer *io, const unsigned char *s)
{
	while (*s) {
		if (*s==255)
			if (-1==buffer_put(io,s,1)) return -1;
		if (-1==buffer_put(io,s++,1)) return -1;
	}
	return 0;
}

int
ftp_cc_write_cmd_s(buffer *io, const unsigned char *s1)
{
	if (ftp_cc_write_escaped(io,s1)) return -1;
	if (-1==buffer_puts(io,"\r\n")) return -1;
	return buffer_flush(io);
}

int
ftp_cc_write_cmd_ss(buffer *io, const unsigned char *s1, 
	const unsigned char *s2)
{
	if (-1==ftp_cc_write_escaped(io,s1)) return -1;
	if (-1==ftp_cc_write_escaped(io,s2)) return -1;
	if (-1==buffer_puts(io,"\r\n")) return -1;
	return buffer_flush(io);
}

/* should probably handle escapes more clever. */
int
ftp_cc_read(buffer *io, stralloc *ret) 
{
	char *p;
	int multiline=0;
	unsigned char *q,*w,*e;
	unsigned int l;
	while (1) {
		int gotlf;
		if (-1==getln(io,ret,&gotlf,'\n')) return -1;
		if (ret->len==0) break;

		p=ret->s;
		l=ret->len;
		for (q=p,e=p+l,w=p;q!=e;) {
			if (*q==255 && q!=e-1) {
				q++;
			}
			*w++=*q++;
		}
		l=w-(unsigned char *)p;
		if (l<2 || p[l-1]!='\n') return 0;
		l--; if (p[l-1]=='\r') l--;
		p[l]=0;
		if (l>=4 
			&& (p[0]>='0' && p[0]<='9')
			&& (p[1]>='0' && p[1]<='9')
			&& (p[2]>='0' && p[2]<='9')
			&& (p[3]==' '))
			break;
		if (l>=4 
			&& (p[0]>='0' && p[0]<='9')
			&& (p[1]>='0' && p[1]<='9')
			&& (p[2]>='0' && p[2]<='9')
			&& (p[3]=='-')) {
			multiline=1;
		}
		if (!multiline) break;
	}
	return 1;
}

int
ftp_cc_pasv(buffer *io_i, buffer *io_o)
{
	stralloc x=STRALLOC_INIT;
	char *p;
	char *q;
	char *r;
	int i;
	unsigned long p1=0,p2=0; /* keep compiler quiet */
	if (-1==ftp_cc_write_cmd_s(io_o,"PASV")) return -1;
	switch(ftp_cc_read(io_i,&x)) {
	case -1: return -1;
	case 0: return 0;
	}

	if (x.len<15) return -2;
	p=x.s;
	q=p+4;
	while (*q && (*q<'0'  ||  *q>'9')) q++;
	if (!*q) {stralloc_free(&x); return -2; }
	for (i=0,r=q;*r;r++) {
		if (*r==',') {
			if (i<3) {
				*r='.'; 
				i++;
			}  else {
				size_t l;
				unsigned long y;
				*r='\0';
				l=str2ulong(r+1,&y,10);
				if (l<=0 || (i==3 && r[l+1]!=',')) 
					{stralloc_free(&x); return -2; }
				if (i++==3) p1=y;
				else p2=y;
			}
		}
	}
	/* cannot overflow the input buffer: even 3,4 is longer than 12 */
	i=str_len(q);
	q[i++]=':';
	i+=fmt_ulong(q+i,p1*256+p2);
	q[i]=0;

	i=doconnect(q);
	stralloc_free(&x);
	return i;
}

