#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/time.h>
#include "_stralloc.h"
#include "ftpparse.h"
#include "_buffer.h"
#include "ftplib.h"
#include "str_ulong.h"
#include "uogetopt.h"
#include "uotime.h"
#include "doconnect.h"
#include "urlparse.h"
#include "timeoutio.h"
#include "getln.h"
#include "fmt.h"
#include "_str.h"
#include "error.h"
#include "ftpcopy.h"
#include "bailout.h"

buffer io_i;
buffer io_o;
int data_sock=-1;
static const char *o_user;
static const char *o_pass;
static const char *o_acct;
static const char *o_title;
static const char *o_urlprefix;
unsigned long o_loglevel=1;
static int o_html=0;
static int o_machine=0;
enum { SM_NEW, SM_OLD, SM_SHORT, SM_LONG, SM_ALPHA, SM_OMEGA } sortmode_e;
static unsigned long sortmode=SM_NEW;
static const char *o_list_options=0;
static stralloc urlprefix;
static unsigned long o_tries=1;


#define KB (1024)
#define MB (1024*1024)
static void mksize(stralloc *out,size_t l)
{
	char buf[STR_ULONG];
	if (l<10000) {
		str_ulong_fmt(buf,l,10,4,' ');
		if (!stralloc_cats(out,buf)) oom();
		if (!stralloc_cats(out,"    B")) oom();
	}
	else if (l<9*MB) {
		str_ulong_fmt(buf,l/KB,10,4,' ');
		if (!stralloc_cats(out,buf)) oom();
		if (!stralloc_cats(out,".")) oom();

		str_ulong_fmt(buf,(l-(l/KB)*KB)/102,10,1,' ');
		if (!stralloc_cats(out,buf)) oom();
		if (!stralloc_cats(out," KB")) oom();
	} else {
		str_ulong_fmt(buf,l/MB,10,6,' ');
		if (!stralloc_cats(out,buf)) oom();
		if (!stralloc_cats(out,".")) oom();
		str_ulong_fmt(buf,(l-(l/MB)*MB)/(102*KB),10,1,' ');
		if (!stralloc_cats(out,buf)) oom();
		if (!stralloc_cats(out," MB")) oom();
	}
}

static void
entity(struct ftpparse *x)
{
	char nb[STR_ULONG];
	char db[20];
	static stralloc sizebuf=STRALLOC_INIT;
	uo_sec70_t s70;
	uo_datetime_t dt;
	s70=x->mtime;
	uo_sec702dt(&dt,&s70);
	yyyy_mm_dd_hh_mm_ss(db,&dt,'-');
	if (o_machine) {
		log1("N");
		logmem(x->name,x->namelen);
		log1("\nK");
		if (x->flagtrycwd && x->flagtryretr) log1("link"); /* guess */
		else if (x->flagtrycwd) log1("dir");
		else log1("file");
		log1("\nS");
		nb[fmt_ulong(nb,x->size)]=0;
		log4(nb,"\nT",db,"\n");
	} else if (o_html) {
		log1("<dt><a href=\"");
		logmem(urlprefix.s,urlprefix.len);
		logmem(x->name,x->namelen);
		log1("\">");
		logmem(x->name,x->namelen);
		log1("</a><br>\n");
		log3("<dd> last modified ",db,", ");
		sizebuf.len=0;
		mksize(&sizebuf,x->size);
		logmem(sizebuf.s,sizebuf.len);
		log1("\n");
	} else{
		if (x->flagtrycwd && x->flagtryretr) log1("link "); /* guess */
		else if (x->flagtrycwd) log1("dir  ");
		else log1("file ");
		log2(db,"  ");
		sizebuf.len=0;
		mksize(&sizebuf,x->size);
		logmem(sizebuf.s,sizebuf.len);
		log1(" ");
		logmem(x->name,x->namelen);
		log1("\n");
	}
}

static stralloc dirdata=STRALLOC_INIT;
static int dirlines=0;

static int 
get_listing(stralloc *r_dir)
{
	char *p;

	if (data_sock==-1)
		data_sock=do_pasv();

	cmdwrite2("CWD ",r_dir->s);

	p=ccread();
	if (!p) bailout(errno,"read (CWD answer)",0,0,0);
	if (*p=='2') {
		if (o_list_options)
			cmdwrite2("LIST ", o_list_options);
		else
			cmdwrite1("LIST");
		p=ccread();
		if (!p) bailout(errno,"read (LIST answer)",0,0,0);
		if (*p!='1')
			return 0;
	}
	if (*p!='1') {
		cmdwrite2("LIST ",r_dir->s);
		p=ccread();
		if (!p) bailout(errno,"read (LIST answer)",0,0,0);
		if (*p!='1')
			return 0;
	}

	dirlines=ftp_read_list(data_sock,&dirdata);
	if (-1==dirlines)
		bailout(errno,"cannot read directory",0,0,0);
	close(data_sock);
	data_sock=-1;
	x2("LIST");
	return 1;
}

static int cmp_them(const void *a, const void *b)
{
	const struct ftpparse *fa=a;
	const struct ftpparse *fb=b;
	if (sortmode==SM_NEW) return fb->mtime-fa->mtime;
	if (sortmode==SM_OLD) return fa->mtime-fb->mtime;
	if (sortmode==SM_LONG) return fb->size-fa->size;
	if (sortmode==SM_SHORT) return fa->size-fb->size;
	if (fa->namelen==fb->namelen)	
		return memcmp(fa->name,fb->name,fa->namelen)
			* (sortmode==SM_ALPHA? 1 : -1);
	else {
		int x;
		x = memcmp (fa->name, fb->name, fa->namelen < fb->namelen ?
				fa->namelen : fb->namelen);
		if (x) return x * (sortmode==SM_ALPHA? 1 : -1);
		if (fa->namelen<fb->namelen) return 1 * (sortmode==SM_ALPHA? 1 :-1);
		if (fa->namelen==fb->namelen) return 0;
		return -1 * (sortmode==SM_ALPHA? 1 :-1);
	}
}

static void
sort_it(struct ftpparse *fp, int count)
{
	qsort(fp,count,sizeof(*fp),cmp_them);
}

static int
parseit(stralloc *r_dir)
{
	char *p,*e;
	int ind;
	struct ftpparse *parsed;
	p=dirdata.s;
	e=dirdata.s+dirdata.len;
	parsed=malloc(sizeof(*parsed) * dirlines);
	if (!parsed) oom();
	ind=0;

	while (p!=e) {
		size_t l=strlen(p);
		if (0==ftpparse(&parsed[ind],p,l)) {
			if (!str_start(p,"total") && !str_start(p,"Total")) 
				log3("cannot parse LIST line: ",p,"\r\n");
		} else if (parsed[ind].name[0]=='.'
			&& (parsed[ind].namelen==1 || (parsed[ind].namelen==2 && parsed[ind].name[1]=='.'))) {
			if (o_loglevel > 1) {
				log2(r_dir->s,"/");
				logmem(parsed[ind].name,parsed[ind].namelen);
				log1(": ignored\n");
			}
		} else {
			ind++;
		}
		p+=l+1;
	}
	sort_it(parsed,ind);
	{
		int i;
		for (i=0;i<ind;i++) {
			entity(&parsed[i]);
		}
	}
	return 1;
}

static uogetopt_t myopts[] =
{
{  0, "account", UOGO_STRING, &o_acct,0,
"Send ACCOUNT as account name during login phase.\n"
"Note: this is _not_ the user name (-u), but the\n"
"name of what could be called a subaccount\n"
"implemented by a few servers.\n","ACCOUNT"},
{'a', "alpha", UOGO_FLAG, &sortmode, SM_ALPHA, "sort alphabetical.",0},
{'z', "omega", UOGO_FLAG, &sortmode, SM_OMEGA, "sort reverse alphabetical.",0},
{'n', "newest", UOGO_FLAG, &sortmode, SM_NEW, "sort youngest first.",0},
{'o', "oldest", UOGO_FLAG, &sortmode, SM_OLD, "sort oldest first.",0},
{'s', "shortest", UOGO_FLAG, &sortmode, SM_SHORT, "sort shortest first.",0},
{'l', "longest", UOGO_FLAG, &sortmode, SM_LONG, "sort longest first.",0},
{'L', "list-options", UOGO_STRING, &o_list_options,1,
"Add OPTS to LIST command.\n"
"This allows to pass arbitrary options to the\n"
"FTP servers LIST command. Note that ftpcopy does\n"
"not cope well with recursive directory listings.","OPTS"},
{'h', "html", UOGO_FLAG, &o_html, 1, "Create HTML index.\n",0},
{'m', "machine", UOGO_FLAG, &o_machine, 1, "Create machine parsable output.",0},
{'t', "title", UOGO_STRING, &o_title,0,
"Title text to use on the HTML output.","TEXT"},
{'\0', "tries", UOGO_ULONG, &o_tries,0,
"Number of tries to connect and log in.\n"
"The default is 1.",0},
{'p', "pass", UOGO_STRING, &o_pass,0,
"Use PASS as password to login on the ftp server.\n"
"The default is `anonymous@invalid.example'.","PASSWORD"},
{'u', "user", UOGO_STRING, &o_user,0,
"Use NAME to login on the ftp server.\n"
"The default is `anonymous'. Use an empty name\n"
"to force to not log in.","NAME"},
{'U', "urlprefix", UOGO_STRING, &o_urlprefix,0,
"URL-Prefix to use in listings.\n"
"The default is taken from the arguments.","URL"},
{0,"",UOGO_MINARGS | UOGO_HIDDEN,0, 2 , 0,0},
{0,"",UOGO_MAXARGS | UOGO_HIDDEN,0, 3 , 0,0},
{0, 0, 0, 0, 0, 0, 0}           /* --help and --version */
};

static void 
usage(void) 
{
	bailout(0,
	  "usage: ftpls [options] host[:port] remotedir\n"
	  "   or: ftpls [options] URL\n"
	  "  use the --help option to get a description of the options",
	  0,0,0);
}

int 
main(int argc, char **argv)
{
	const char *host,*remotedir;
	stralloc d1=STRALLOC_INIT;
	stralloc proto=STRALLOC_INIT;
	stralloc user=STRALLOC_INIT;
	stralloc pass=STRALLOC_INIT;
	stralloc hostport=STRALLOC_INIT;
	stralloc rest=STRALLOC_INIT;

	bailout_progname(argv[0]);

	uogetopt (flag_bailout_log_name, PACKAGE, VERSION, 
		&argc, argv, uogetopt_out, 
	  "usage: ftpls [options] host[:port] remotedir\n"
	  "   or: ftpls [options] URL\n",
	  myopts,
	  "\nReport bugs to <uwe-ftpcopy@bulkmail.ohse.de>");
	if (urlparse(argv[1],&proto,&user,&pass,&hostport,&rest)) {
		if (!stralloc_0(&proto)) oom();
		if (!str_equal(proto.s,"ftp")) 
			bailout(0,"unsupported URL type",0,0,0);
		if (!hostport.len) bailout(0,"empty host in url",0,0,0);
		if (!stralloc_0(&hostport)) oom();
		host=hostport.s;
		if (!rest.len) if (!stralloc_append(&rest,"/")) oom();
		if (!stralloc_0(&rest)) oom();
		remotedir=rest.s;
		if (!o_user && user.len) {
			if (!stralloc_0(&user)) oom();
			o_user=user.s;
		}
		if (!o_pass && pass.len) {
			if (!stralloc_0(&pass)) oom();
			o_user=pass.s;
		}
		if (argv[2]) usage();

		if (!stralloc_copys(&urlprefix,argv[1])) oom();
		if (urlprefix.s[urlprefix.len-1]!='/')
			if (!stralloc_append(&urlprefix,"/")) oom();
	} else {
		host=argv[1];
		if (!argv[2]) 
			usage();
		remotedir=argv[2];
		if (o_html) {
			if (!stralloc_copys(&urlprefix,"ftp://")) oom();
			if (!stralloc_cats(&urlprefix,host)) oom();
			if (!stralloc_cats(&urlprefix,remotedir)) oom();
			if (urlprefix.s[urlprefix.len-1]!='/')
				if (!stralloc_append(&urlprefix,"/")) oom();
		} else
			if (!stralloc_0(&urlprefix)) oom();
	}
	if (!o_user) o_user="anonymous";
	if (!o_pass) o_pass="anonymous@example.invalid";
	if (o_urlprefix) if (!stralloc_copys(&urlprefix,o_urlprefix)) oom();

	if (!stralloc_copys(&d1,remotedir) || !stralloc_0(&d1)) oom();

	connect_auth(host,o_user,o_pass,o_acct,o_tries);

	sx2("TYPE I");
	if (!get_listing(&d1))
		exit(1);
	if (o_html) {
		log1("<html><head>\n");
		if (o_title) log3("<title>",o_title,"</title>\n");
		log1("</head><body>\n");
		if (o_title) {
			log3("<h1>",o_title,"</h1>\n");
		}
		log1("<dl>\n");
	}
	parseit(&d1);
	if (o_html) {
		log1("</dl></body></html>\n");
	}
	exit(0);
}
